#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import logging
import typing

from axosyslog_light.common.file import File
from axosyslog_light.common.operations import cast_to_list
from axosyslog_light.syslog_ng_config import stringify
from axosyslog_light.syslog_ng_config.renderer import ConfigRenderer
from axosyslog_light.syslog_ng_config.statement_group import StatementGroup
from axosyslog_light.syslog_ng_config.statements import ArrowedOptions
from axosyslog_light.syslog_ng_config.statements.destinations.example_destination import ExampleDestination
from axosyslog_light.syslog_ng_config.statements.destinations.file_destination import FileDestination
from axosyslog_light.syslog_ng_config.statements.destinations.network_destination import NetworkDestination
from axosyslog_light.syslog_ng_config.statements.destinations.snmp_destination import SnmpDestination
from axosyslog_light.syslog_ng_config.statements.destinations.unix_dgram_destination import UnixDgramDestination
from axosyslog_light.syslog_ng_config.statements.destinations.unix_stream_destination import UnixStreamDestination
from axosyslog_light.syslog_ng_config.statements.filters.filter import Filter
from axosyslog_light.syslog_ng_config.statements.filters.filter import Match
from axosyslog_light.syslog_ng_config.statements.filters.filter import RateLimit
from axosyslog_light.syslog_ng_config.statements.filterx.filterx import FilterX
from axosyslog_light.syslog_ng_config.statements.logpath.log_context_scl_block import LogContextSclBlock
from axosyslog_light.syslog_ng_config.statements.logpath.logpath import LogPath
from axosyslog_light.syslog_ng_config.statements.parsers.db_parser import DBParser
from axosyslog_light.syslog_ng_config.statements.parsers.parser import Parser
from axosyslog_light.syslog_ng_config.statements.rewrite.rewrite import CreditCardHash
from axosyslog_light.syslog_ng_config.statements.rewrite.rewrite import CreditCardMask
from axosyslog_light.syslog_ng_config.statements.rewrite.rewrite import Set
from axosyslog_light.syslog_ng_config.statements.rewrite.rewrite import SetPri
from axosyslog_light.syslog_ng_config.statements.rewrite.rewrite import SetTag
from axosyslog_light.syslog_ng_config.statements.sources.example_msg_generator_source import ExampleMsgGeneratorSource
from axosyslog_light.syslog_ng_config.statements.sources.file_source import FileSource
from axosyslog_light.syslog_ng_config.statements.sources.internal_source import InternalSource
from axosyslog_light.syslog_ng_config.statements.sources.network_source import NetworkSource
from axosyslog_light.syslog_ng_config.statements.sources.opentelemetry_source import OpenTelemetrySource
from axosyslog_light.syslog_ng_config.statements.sources.syslog_source import SyslogSource
from axosyslog_light.syslog_ng_config.statements.sources.unix_dgram_source import UnixDgramSource
from axosyslog_light.syslog_ng_config.statements.sources.webhook_source import WebhookJsonSource
from axosyslog_light.syslog_ng_config.statements.sources.webhook_source import WebhookSource
from axosyslog_light.syslog_ng_config.statements.template.template import Template
from axosyslog_light.syslog_ng_config.statements.template.template import TemplateFunction
from axosyslog_light.syslog_ng_ctl.legacy_stats_handler import LegacyStatsHandler
from axosyslog_light.syslog_ng_ctl.prometheus_stats_handler import MetricFilter
from axosyslog_light.syslog_ng_ctl.prometheus_stats_handler import PrometheusStatsHandler
from axosyslog_light.syslog_ng_ctl.prometheus_stats_handler import Sample


logger = logging.getLogger(__name__)


class SyslogNgConfig(object):
    def __init__(
        self,
        config_version: str,
        stats_handler: LegacyStatsHandler,
        prometheus_stats_handler: PrometheusStatsHandler,
        teardown,
    ) -> None:
        self._raw_config = None
        self._syslog_ng_config = {
            "version": config_version,
            "includes": [],
            "global_options": {},
            "preamble": "",
            "templates": [],
            "statement_groups": [],
            "logpath_groups": [],
        }
        self._stats_handler = stats_handler
        self._prometheus_stats_handler = prometheus_stats_handler
        self.__implicit_statement_groups: typing.Dict[int, StatementGroup] = {}
        self.teardown = teardown

    stringify = staticmethod(stringify)

    @property
    def stats_handler(self) -> LegacyStatsHandler:
        return self._stats_handler

    @property
    def prometheus_stats_handler(self) -> PrometheusStatsHandler:
        return self._prometheus_stats_handler

    def arrowed_options(self, *args, **kwargs):
        return ArrowedOptions(*args, **kwargs)

    def set_raw_config(self, raw_config):
        self._raw_config = raw_config

    def write_config(self, config_path):
        if self._raw_config:
            rendered_config = self._raw_config
        else:
            rendered_config = ConfigRenderer(self._syslog_ng_config).get_rendered_config()
        logger.info("Generated syslog-ng config\n{}\n".format(rendered_config))

        syslog_ng_config_file = File(config_path)
        syslog_ng_config_file.write_content_and_close(rendered_config)

    def set_version(self, version):
        self._syslog_ng_config["version"] = version

    def get_version(self):
        return self._syslog_ng_config["version"]

    def add_include(self, include):
        self._syslog_ng_config["includes"].append(include)

    def update_global_options(self, **options):
        self._syslog_ng_config["global_options"].update(options)

    def add_preamble(self, preamble):
        self._syslog_ng_config["preamble"] += "\n" + preamble

    def add_template(self, template):
        self._syslog_ng_config["templates"].append(template)

    def create_file_source(self, **options):
        file_source = FileSource(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(file_source.close_file)
        return file_source

    def create_unix_dgram_source(self, **options):
        unix_dgram_source = UnixDgramSource(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(unix_dgram_source.close_socket)
        return unix_dgram_source

    def create_example_msg_generator_source(self, **options):
        return ExampleMsgGeneratorSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_internal_source(self, **options):
        return InternalSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_network_source(self, **options):
        return NetworkSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_syslog_source(self, **options):
        return SyslogSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_opentelemetry_source(self, **options):
        return OpenTelemetrySource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_webhook_source(self, **options):
        return WebhookSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_webhook_json_source(self, **options):
        return WebhookJsonSource(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_rewrite_set(self, template, **options):
        return Set(template, **options)

    def create_rewrite_set_tag(self, tag, **options):
        return SetTag(tag, **options)

    def create_rewrite_set_pri(self, pri, **options):
        return SetPri(pri, **options)

    def create_template(self, template, **options):
        return Template(template, **options)

    def create_template_function(self, template, **options):
        return TemplateFunction(template, **options)

    def create_filterx(self, code: str):
        return FilterX(code)

    def create_filter(self, expr=None, **options):
        return Filter("", self._stats_handler, self._prometheus_stats_handler, [expr] if expr else [], **options)

    def create_rate_limit_filter(self, **options):
        return RateLimit(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_match_filter(self, **options):
        return Match(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_app_parser(self, **options):
        return Parser("app-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_checkpoint_parser(self, **options):
        return Parser("checkpoint-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_panos_parser(self, **options):
        return Parser("panos-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_regexp_parser(self, **options):
        return Parser("regexp-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_csv_parser(self, **options):
        return Parser("csv-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_metrics_probe(self, **options):
        return Parser("metrics_probe", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_syslog_parser(self, **options):
        return Parser("syslog-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_sdata_parser(self, **options):
        return Parser("sdata-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_group_lines_parser(self, **options):
        return Parser("group-lines", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_cisco_parser(self, **options):
        return Parser("cisco-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_mariadb_audit_parser(self, **options):
        return Parser("mariadb-audit-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_postgresql_csvlog_parser(self, **options):
        return Parser("postgresql-csvlog-parser", self._stats_handler, self._prometheus_stats_handler, **options)

    def create_file_destination(self, **options):
        file_destination = FileDestination(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(file_destination.close_file)
        return file_destination

    def create_example_destination(self, **options):
        example_destination = ExampleDestination(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(example_destination.close_file)
        return example_destination

    def create_snmp_destination(self, **options):
        return SnmpDestination(self._stats_handler, self._prometheus_stats_handler, **options)

    def create_network_destination(self, **options):
        network_destination = NetworkDestination(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(network_destination.stop_listener)
        return network_destination

    def create_unix_dgram_destination(self, **options):
        unix_dgram_destination = UnixDgramDestination(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(unix_dgram_destination.stop_listener)
        return unix_dgram_destination

    def create_unix_stream_destination(self, **options):
        unix_stream_source = UnixStreamDestination(self._stats_handler, self._prometheus_stats_handler, **options)
        self.teardown.register(unix_stream_source.stop_listener)
        return unix_stream_source

    def create_db_parser(self, config, **options):
        return DBParser(self._stats_handler, self._prometheus_stats_handler, config, **options)

    def create_rewrite_credit_card_mask(self, **options):
        return CreditCardMask(**options)

    def create_rewrite_credit_card_hash(self, **options):
        return CreditCardHash(**options)

    def create_logpath(self, name=None, statements=None, flags=None):
        logpath = self.__create_logpath_with_conversion(name, statements, flags)
        self._syslog_ng_config["logpath_groups"].append(logpath)
        return logpath

    def create_inner_logpath(self, name=None, statements=None, flags=None):
        inner_logpath = self.__create_logpath_with_conversion(name, statements, flags)
        return inner_logpath

    def create_statement_group(self, statements):
        statement_group = StatementGroup(statements)
        self._syslog_ng_config["statement_groups"].append(statement_group)
        return statement_group

    def __create_statement_group_if_needed(self, item):
        if isinstance(item, (StatementGroup, LogPath, FilterX, LogContextSclBlock)):
            return item
        else:
            item_id = id(item)
            group = self.__implicit_statement_groups.get(item_id)
            if group is None:
                group = self.__implicit_statement_groups[item_id] = self.create_statement_group(item)
            return group

    def __create_logpath_with_conversion(self, name, items, flags):
        return self.__create_logpath_group(
            name,
            map(self.__create_statement_group_if_needed, cast_to_list(items)),
            flags,
        )

    def __create_logpath_group(self, name=None, statements=None, flags=None):
        logpath = LogPath(self._prometheus_stats_handler, name)
        if statements:
            logpath.add_groups(statements)
        if flags:
            logpath.add_flags(cast_to_list(flags))
        return logpath

    def get_prometheus_samples(self, metric_filter: typing.List[MetricFilter]) -> typing.List[Sample]:
        return self.prometheus_stats_handler.get_samples(metric_filter)



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2020 One Identity
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import typing
from pathlib import Path

from axosyslog_light.common.session_data import get_session_data
from axosyslog_light.helpers.loggen.loggen_executor import LoggenExecutor
from axosyslog_light.helpers.loggen.loggen_executor import LoggenStartParams
from psutil import Popen


class LoggenStats:
    def __init__(self, loggen_stderr_file: Path) -> None:
        loggen_stderr_last_line = loggen_stderr_file.read_text(encoding="utf8").splitlines()[-1].strip()
        self.msg_per_second = float(loggen_stderr_last_line.split("avg_rate=")[1].split(" msg/s")[0])
        self.stdev_rate = float(loggen_stderr_last_line.split("stdev_rate=")[1].split("%")[0].replace(",", "."))
        self.total_msg_count = int(loggen_stderr_last_line.split("count=")[1].split(",")[0])
        self.run_time = float(loggen_stderr_last_line.split("time=")[1].split(",")[0])
        self.avg_msg_size = int(loggen_stderr_last_line.split("avg_msg_size=")[1].split(",")[0])
        self.bandwidth = float(loggen_stderr_last_line.split("bandwidth=")[1].split(" KiB/s")[0])


class Loggen(object):
    @staticmethod
    def __get_new_instance_index():
        with get_session_data() as session_data:
            last_index = session_data.get("loggen_instance_index", 0)
            index = last_index + 1
            session_data["loggen_instance_index"] = index
        return index

    def __init__(self, loggen_executor: typing.Optional[LoggenExecutor] = None):
        if loggen_executor is None:
            loggen_executor = LoggenExecutor.get_default_executor()

        if not loggen_executor:
            raise Exception("Loggen executor is not available")

        self.executor = loggen_executor

    def start(self, start_params: LoggenStartParams) -> Popen:
        instance_index = Loggen.__get_new_instance_index()
        self.loggen_stdout_path = Path("loggen_stdout_{}".format(instance_index))
        self.loggen_stderr_path = Path("loggen_stderr_{}".format(instance_index))
        return self.executor.start(start_params, self.loggen_stderr_path, self.loggen_stdout_path, instance_index)

    def get_loggen_stats(self) -> LoggenStats:
        if not self.loggen_stderr_path.exists():
            raise RuntimeError("Loggen stderr file does not exist")

        return LoggenStats(self.loggen_stderr_path)

    def stop(self) -> None:
        self.executor.stop()

    def get_sent_message_count(self) -> int:
        if not self.loggen_stderr_path.exists():
            return 0

        # loggen puts the count= messages to the stderr
        f = open(self.loggen_stderr_path, "r")
        content = f.read()
        f.close()

        start_pattern = "count="
        if start_pattern not in content:
            return 0
        index_start = content.rindex(start_pattern) + len(start_pattern)

        end_pattern = ", "
        if end_pattern not in content:
            return 0
        index_end = content.find(end_pattern, index_start)

        return int(content[index_start:index_end])



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2020 One Identity
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
from pathlib import Path

import pytest
from axosyslog_light.common.file import copy_file
from axosyslog_light.executors.command_executor import CommandExecutor
from axosyslog_light.syslog_ng.syslog_ng_paths import SyslogNgPaths


class SecureLogging():
    def __init__(self, testcase_parameters):
        self.instance_paths = SyslogNgPaths(testcase_parameters).set_syslog_ng_paths("server")
        self.master_key = None
        self.derived_key = None
        self.decryption_key = None
        self.slogkey = self.instance_paths.get_slogkey_bin()
        self.slogverify = self.instance_paths.get_slogverify_bin()

        self.create_master_key()
        self.create_derived_key()
        self.create_decryption_key()

    def create_master_key(self):
        slogkey_stdout = Path("slogkey_stdout_master")
        slogkey_stderr = Path("slogkey_stderr_master")

        self.master_key = Path("master.key")

        CommandExecutor().run(
            [self.slogkey, "-m", self.master_key],
            slogkey_stdout,
            slogkey_stderr,
        )

    def create_derived_key(self):
        slogkey_stdout = Path("slogkey_stdout_derived")
        slogkey_stderr = Path("slogkey_stderr_derived")

        self.derived_key = Path("derived.key")
        self.cmac = Path("cmac")

        CommandExecutor().run(
            [self.slogkey, "-d", self.master_key, "foo", "bar", self.derived_key],
            slogkey_stdout,
            slogkey_stderr,
        )

    def create_decryption_key(self):
        self.decryption_key = Path("decryption.key")
        copy_file(self.derived_key, self.decryption_key)

    def decrypt(self, input_file):
        slogverify_stdout = Path("slogverify_stdout")
        slogverify_stderr = Path("slogverify_stderr")
        encrypted = Path(input_file)
        decrypted = Path("decrypted.txt")

        CommandExecutor().run(
            [
                self.slogverify,
                "-k", self.decryption_key,
                "-m", self.cmac,
                encrypted, decrypted,
            ],
            slogverify_stdout,
            slogverify_stderr,
        )
        return decrypted.read_text().rstrip("\n").split("\n")


@pytest.fixture
def slog(testcase_parameters):
    yield SecureLogging(testcase_parameters)



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import pytest
from axosyslog_light.executors.command_executor import CommandExecutor


@pytest.mark.parametrize(
    "command, expected_stdout, expected_stderr, expected_exit_code",
    [(["grep", "a", "a"], "", "grep: a: No such file or directory\n", 2), (["echo", "a"], "a\n", "", 0)],
)
def test_execute_command(tmpdir, command, expected_stdout, expected_stderr, expected_exit_code):
    stdout_file = tmpdir.join("stdout.log")
    stderr_file = tmpdir.join("stderr.log")
    command_executor = CommandExecutor()
    result = command_executor.run(command, stdout_file, stderr_file)
    assert result["stdout"] == expected_stdout
    assert result["stderr"] == expected_stderr
    assert result["exit_code"] == expected_exit_code



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
from axosyslog_light.syslog_ng.syslog_ng_paths import SyslogNgPaths
from axosyslog_light.syslog_ng_ctl.syslog_ng_ctl_executor import QueryTypes
from axosyslog_light.syslog_ng_ctl.syslog_ng_ctl_executor import SyslogNgCtlExecutor


class SyslogNgCtl(object):
    def __init__(self, instance_paths: SyslogNgPaths, syslog_ng_ctl_executor: SyslogNgCtlExecutor) -> None:
        self.__instance_name = instance_paths.get_instance_name()
        self.__syslog_ng_ctl_executor = syslog_ng_ctl_executor

    def reload(self):
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="reload",
            command=["reload"],
        )

    def stop(self):
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="stop",
            command=["stop"],
        )

    def stats(self, reset=False):
        ctl_stats_command = self.__syslog_ng_ctl_executor.construct_ctl_stats_command(reset=reset)
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="stats",
            command=ctl_stats_command,
        )

    def stats_prometheus(self):
        ctl_stats_prometheus_command = self.__syslog_ng_ctl_executor.construct_ctl_stats_prometheus_command()
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="stats_prometheus",
            command=ctl_stats_prometheus_command,
        )

    def query(self, pattern="*", query_type=QueryTypes.QUERY_GET):
        ctl_query_command = self.__syslog_ng_ctl_executor.construct_ctl_query_command(pattern, query_type)
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="query",
            command=ctl_query_command,
        )

    def credentials_add(self, credential, secret):
        ctl_credentials_command = self.__syslog_ng_ctl_executor.construct_ctl_credentials_command(credential, secret)
        return self.__syslog_ng_ctl_executor.run_command(
            self.__instance_name,
            command_short_name="credentials",
            command=ctl_credentials_command,
        )

    def is_control_socket_alive(self):
        return self.stats(reset=False)["exit_code"] == 0



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import logging
import typing
from copy import copy
from pathlib import Path
from subprocess import Popen

from axosyslog_light.common.blocking import wait_until_false
from axosyslog_light.common.blocking import wait_until_true
from axosyslog_light.syslog_ng.console_log_reader import ConsoleLogReader
from axosyslog_light.syslog_ng.syslog_ng_executor import SyslogNgExecutor
from axosyslog_light.syslog_ng.syslog_ng_executor import SyslogNgStartParams
from axosyslog_light.syslog_ng.syslog_ng_paths import SyslogNgPaths
from axosyslog_light.syslog_ng_config.syslog_ng_config import SyslogNgConfig
from axosyslog_light.syslog_ng_ctl.legacy_stats_handler import LegacyStatsHandler
from axosyslog_light.syslog_ng_ctl.prometheus_stats_handler import PrometheusStatsHandler
from axosyslog_light.syslog_ng_ctl.syslog_ng_ctl import SyslogNgCtl
from axosyslog_light.testcase_parameters.testcase_parameters import TestcaseParameters

logger = logging.getLogger(__name__)


class SyslogNg(object):
    def __init__(
        self,
        syslog_ng_executor: SyslogNgExecutor,
        syslog_ng_ctl: SyslogNgCtl,
        instance_paths: SyslogNgPaths,
        testcase_parameters: TestcaseParameters,
        teardown,
    ) -> None:
        self.instance_paths = instance_paths
        self.start_params = SyslogNgStartParams(
            config_path=self.instance_paths.get_config_path(),
            persist_path=self.instance_paths.get_persist_path(),
            pid_path=self.instance_paths.get_pid_path(),
            control_socket_path=self.instance_paths.get_control_socket_path(),
        )
        self._external_tool = testcase_parameters.get_external_tool()
        self._console_log_reader = ConsoleLogReader(self.instance_paths, teardown)
        self._syslog_ng_ctl = syslog_ng_ctl
        self._syslog_ng_executor = syslog_ng_executor
        self._process: typing.Optional[Popen] = None

    def create_config(self, config_version: str, teardown) -> SyslogNgConfig:
        return SyslogNgConfig(
            config_version,
            LegacyStatsHandler(self._syslog_ng_ctl),
            PrometheusStatsHandler(self._syslog_ng_ctl),
            teardown,
        )

    def start(self, config: SyslogNgConfig) -> Popen:
        if self._process:
            raise Exception("syslog-ng has been already started")

        config.write_config(self.instance_paths.get_config_path())

        self.__syntax_check()
        self.__start_syslog_ng()

        logger.info("syslog-ng process has been started with PID: {}\n".format(self._process.pid))

        return self._process

    def stop(self, unexpected_messages: typing.List[str] = None) -> None:
        if self.is_process_running():
            saved_pid = self._process.pid
            # effective stop
            result = self._syslog_ng_ctl.stop()

            # wait for stop and check stop result
            self.__validate_returncode(result["exit_code"])
            if not wait_until_false(self.is_process_running):
                self.__error_handling("syslog-ng did not stop")
            if self.start_params.stderr and self.start_params.debug and self.start_params.verbose:
                if not self._console_log_reader.wait_for_stop_message():
                    logger.warning("Stop message has not been found, this might be because of a very long console log")
            self._console_log_reader.check_for_unexpected_messages(unexpected_messages)
            if self._external_tool == "valgrind":
                self._console_log_reader.handle_valgrind_log(Path(f"syslog_ng_{self.instance_paths.get_instance_name()}_valgrind_output"))
            logger.info("syslog-ng process has been stopped with PID: {}\n".format(saved_pid))
        else:
            if self._process is not None:
                self.__validate_returncode(self._process.returncode)
            self._console_log_reader.check_for_unexpected_messages(unexpected_messages)
        self._process = None

    def reload(self, config: SyslogNgConfig) -> None:
        config.write_config(self.instance_paths.get_config_path())

        # effective reload
        result = self._syslog_ng_ctl.reload()

        # wait for reload and check reload result
        self.__validate_returncode(result["exit_code"])
        if not self.__wait_for_control_socket_alive():
            self.__error_handling("Control socket not alive")
        if self.start_params.stderr and self.start_params.debug and self.start_params.verbose:
            if not self._console_log_reader.wait_for_reload_message():
                self.__error_handling("Reload message not arrived")
        logger.info("syslog-ng process has been reloaded with PID: {}\n".format(self._process.pid))

    def restart(self, config: SyslogNgConfig) -> None:
        self.stop()
        self.start(config)

    def _get_version_info(self, keyword: str) -> str:
        stdout_path = Path(f"syslog_ng_{self.instance_paths.get_instance_name()}_version_stdout")
        stderr_path = Path(f"syslog_ng_{self.instance_paths.get_instance_name()}_version_stderr")

        start_params = copy(self.start_params)
        start_params.version = True

        process = self._syslog_ng_executor.run_process(
            start_params=start_params,
            stderr_path=stderr_path,
            stdout_path=stdout_path,
        )
        returncode = process.wait()
        if returncode != 0:
            raise Exception(f"Cannot get syslog-ng {keyword.lower()}. Process returned with {returncode}. See {stderr_path.absolute()} for details")

        for line in stdout_path.read_text().splitlines():
            if keyword in line:
                return line.split(":")[1].lstrip()
        raise Exception(f"Cannot parse '{keyword}' from 'syslog-ng --version'")

    def get_config_version(self) -> str:
        return self._get_version_info("Config version:")

    def get_version(self) -> str:
        return self._get_version_info("Installer-Version:")

    def is_process_running(self) -> bool:
        return self._process and self._process.poll() is None

    def wait_for_messages_in_console_log(self, expected_messages: typing.List[str]) -> typing.List[str]:
        assert issubclass(type(expected_messages), list)
        return self._console_log_reader.wait_for_messages_in_console_log(expected_messages)

    def wait_for_message_in_console_log(self, expected_message: str) -> typing.List[str]:
        return self.wait_for_messages_in_console_log([expected_message])

    def __syntax_check(self) -> None:
        stdout_path = Path(self.instance_paths.get_syntax_only_stdout_path())
        stderr_path = Path(self.instance_paths.get_syntax_only_stderr_path())

        start_params = copy(self.start_params)
        start_params.syntax_only = True

        process = self._syslog_ng_executor.run_process(
            start_params=start_params,
            stderr_path=stderr_path,
            stdout_path=stdout_path,
        )
        returncode = process.wait()
        if returncode == 0:
            logger.info("syslog-ng config syntax check passed")
        elif returncode == 1:
            raise Exception(f"syslog-ng config syntax error. See {stderr_path.absolute()} for details")
        else:
            self.__validate_returncode(returncode)

    def __wait_for_control_socket_alive(self) -> bool:
        def is_alive(s: SyslogNg) -> bool:
            if not s.is_process_running():
                self.__validate_returncode(self._process.returncode)
                self.__error_handling("syslog-ng is not running")
            return s._syslog_ng_ctl.is_control_socket_alive()
        return wait_until_true(is_alive, self)

    def __wait_for_start(self) -> None:
        # wait for start and check start result
        if not self.__wait_for_control_socket_alive():
            self.__error_handling("Control socket not alive")
        if not self._console_log_reader.wait_for_start_message():
            self.__error_handling("Start message not arrived")

    def __start_syslog_ng(self) -> None:
        args = {
            "start_params": self.start_params,
            "stderr_path": self.instance_paths.get_stderr_path(),
            "stdout_path": self.instance_paths.get_stdout_path(),
        }

        if self._external_tool is None:
            self._process = self._syslog_ng_executor.run_process(**args)
        elif self._external_tool == "gdb":
            self._process = self._syslog_ng_executor.run_process_with_gdb(**args)
        elif self._external_tool == "gdb_for_bt":
            self._process = self._syslog_ng_executor.run_process_with_gdb_for_bt(**args)
        elif self._external_tool == "valgrind":
            self._process = self._syslog_ng_executor.run_process_with_valgrind(
                valgrind_output_path=Path(f"syslog_ng_{self.instance_paths.get_instance_name()}_valgrind_output"),
                **args,
            )
        elif self._external_tool == "strace":
            self._process = self._syslog_ng_executor.run_process_with_strace(
                strace_output_path=Path(f"syslog_ng_{self.instance_paths.get_instance_name()}_strace_output"),
                **args,
            )
        else:
            raise Exception(f"Unknown external tool: {self._external_tool}")

        if self.start_params.stderr and self.start_params.debug and self.start_params.verbose:
            if self.start_params.preprocess_into is None and not self.start_params.syntax_only:
                self.__wait_for_start()
            else:
                self._process.wait()

    def __error_handling(self, error_message: str) -> typing.NoReturn:
        self._console_log_reader.dump_stderr()
        raise Exception(error_message)

    def __validate_returncode(self, returncode: int):
        if returncode not in [0, 1, 2]:
            # return code 1 is a directed way of termination (syntax error), it should not handle as a crash
            # return code 2 is a directed way of termination, it should not handle as a crash
            self._process = None
            assert False, "syslog-ng has crashed with return code: %s" % returncode



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
from axosyslog_light.syslog_ng_ctl.legacy_stats_handler import LegacyStatsHandler
from axosyslog_light.syslog_ng_ctl.prometheus_stats_handler import PrometheusStatsHandler


class SourceDriver(object):
    group_type = "source"
    driver_instance = ""

    def __init__(
        self,
        stats_handler: LegacyStatsHandler,
        prometheus_stats_handler: PrometheusStatsHandler,
        positional_parameters=None,
        options=None,
    ) -> None:
        self.stats_handler = stats_handler
        self.prometheus_stats_handler = prometheus_stats_handler

        if positional_parameters is None:
            positional_parameters = []
        self.positional_parameters = positional_parameters
        if options is None:
            options = {}
        self.options = options

    def get_stats(self):
        return self.stats_handler.get_stats(self.group_type, self.driver_name, self.driver_instance)

    def get_query(self):
        return self.stats_handler.get_query(self.group_type, self.driver_name)



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import typing

import prometheus_client.parser
from axosyslog_light.syslog_ng_ctl.syslog_ng_ctl import SyslogNgCtl
from prometheus_client.samples import Sample

__all__ = ["PrometheusStatsHandler", "MetricFilter", "Sample"]


class MetricFilter(typing.NamedTuple):
    name: str
    labels: typing.Optional[typing.Dict[str, str]] = None


class PrometheusStatsHandler(object):
    def __init__(self, syslog_ng_ctl: SyslogNgCtl) -> None:
        self.__syslog_ng_ctl = syslog_ng_ctl

    def __filter_raw_samples(self, metric_filters: typing.List[MetricFilter], raw_samples: str) -> typing.List[Sample]:
        samples = []

        for metric_family in prometheus_client.parser.text_string_to_metric_families(raw_samples):
            for sample in metric_family.samples:
                for metric_filter in metric_filters:
                    if metric_filter.name == sample.name and metric_filter.labels.items() <= sample.labels.items():
                        samples.append(sample)

        return samples

    def get_samples(self, metric_filters: typing.List[MetricFilter]) -> typing.List[Sample]:
        if len(metric_filters) == 0:
            return []

        ctl_output = self.__syslog_ng_ctl.stats_prometheus()
        if ctl_output["exit_code"] != 0:
            return []

        return self.__filter_raw_samples(metric_filters, ctl_output["stdout"])



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
from axosyslog_light.syslog_ng_config import stringify
from axosyslog_light.syslog_ng_config.statements import ArrowedOptions
from axosyslog_light.syslog_ng_config.statements.filters.filter import Filter
from axosyslog_light.syslog_ng_config.statements.template.template import Template
from axosyslog_light.syslog_ng_config.statements.template.template import TemplateFunction


def render_version(version):
    return "@version: {}\n".format(version)


def render_includes(includes):
    include_lines = ['@include "{}"'.format(include) for include in includes]
    return "\n".join(include_lines)


def render_global_options(global_options):
    globals_options_header = "options {\n"
    globals_options_footer = "};\n"

    config_snippet = globals_options_header
    for option_name, option_value in global_options.items():
        config_snippet += "    {}({});\n".format(option_name, option_value)
    config_snippet += globals_options_footer

    return config_snippet


def render_positional_options(positional_parameters):
    config_snippet = ""
    for parameter in positional_parameters:
        config_snippet += "        {}\n".format(str(parameter))
    return config_snippet


def render_options(name, options):

    config_snippet = ""
    config_snippet += "        {}(\n".format(name)
    config_snippet += render_driver_options(options)
    config_snippet += "        )\n"

    return config_snippet


def render_arrowed_options(name, keys):
    config_snippet = "         {}(\n".format(name)
    for key, value in keys.items():
        config_snippet += "        {} => {}\n".format(stringify(key), value)
    config_snippet += "        )\n"

    return config_snippet


def render_template_statement(template):
    if not template.use_simple_statement:
        config_snippet = "template {} {{\n".format(template.name)
        config_snippet += "        template({});\n".format(stringify(template.template))
        if template.template_escape is not None:
            config_snippet += "        template-escape({});\n".format(template.template_escape)
        config_snippet += "};\n"
    else:
        config_snippet = "template {} {};\n".format(template.name, stringify(template.template))
    return config_snippet


def render_template_function(template):
    return "template-function {} {};\n".format(template.name, stringify(template.template))


def render_template(template):
    if isinstance(template, Template):
        return render_template_statement(template)
    elif isinstance(template, TemplateFunction):
        return render_template_function(template)


def render_templates(templates):
    config_snippet = ""
    for template in templates:
        config_snippet += render_template(template)
    return config_snippet


def render_list(name, options):
    config_snippet = ""

    for element in options:
        config_snippet += "        {}({})\n".format(name, element)

    return config_snippet


def render_name_value(name, value):
    return "        {}({})\n".format(name, value)


def render_driver(name, driver):
    return "        {}({})\n".format(name, render_statement(driver))


def render_driver_options(driver_options):
    config_snippet = ""

    for option_name, option_value in driver_options.items():
        if isinstance(option_value, ArrowedOptions):
            config_snippet += render_arrowed_options(option_name, option_value)
        elif isinstance(option_value, dict):
            config_snippet += render_options(option_name, option_value)
        elif (isinstance(option_value, tuple) or isinstance(option_value, list)):
            config_snippet += render_list(option_name, option_value)
        elif isinstance(option_value, Filter):
            config_snippet += render_driver(option_name, option_value)
        else:
            config_snippet += render_name_value(option_name, option_value)

    return config_snippet


def render_statement(statement):
    config_snippet = ""
    config_snippet += "    {} (\n".format(statement.driver_name)
    config_snippet += render_positional_options(statement.positional_parameters)
    config_snippet += render_driver_options(statement.options)
    config_snippet += "    )"

    return config_snippet


def render_statement_groups(statement_groups):
    config_snippet = ""

    for statement_group in statement_groups:
        # statement header
        config_snippet += "\n{} {} {{\n".format(
            statement_group.group_type, statement_group.group_id,
        )

        for statement in statement_group:
            # driver header
            config_snippet += render_statement(statement)

            # driver footer
            config_snippet += ";\n"

            # statement footer
        config_snippet += "};\n"

    return config_snippet


def _indent(string, depth):
    return depth * "    " + string


def render_logpath_groups(logpath_groups, depth=0):
    config_snippet = "\n"

    for logpath_group in logpath_groups:
        if logpath_group.name:
            config_snippet += _indent("log " + logpath_group.name + " {\n", depth)
        else:
            config_snippet += _indent("log {\n", depth)
        for statement_group in logpath_group.logpath:
            if statement_group.group_type == "log":
                config_snippet += render_logpath_groups(logpath_groups=[statement_group], depth=depth + 1)
            elif statement_group.group_type == "filterx":
                config_snippet += _indent("filterx {\n", depth + 1)
                for line in statement_group.code.split("\n"):
                    config_snippet += _indent(line + "\n", depth + 2)
                config_snippet += _indent("};\n", depth + 1)
            elif statement_group.group_type == "log_context_scl_block":
                config_snippet += _indent(statement_group.name + "(\n", depth + 1)
                for line in render_driver_options(statement_group.options).split("\n"):
                    config_snippet += _indent(line + "\n", depth + 2)
                config_snippet += _indent(");\n", depth + 1)
            else:
                config_snippet += _indent("{}({});\n".format(statement_group.group_type, statement_group.group_id), depth + 1)
        if logpath_group.flags:
            config_snippet += _indent("flags({});\n".format("".join(logpath_group.flags)), depth + 1)
        config_snippet += _indent("};\n", depth)

    return config_snippet


class ConfigRenderer(object):
    def __init__(self, syslog_ng_config):
        self.__syslog_ng_config = syslog_ng_config
        self.__syslog_ng_config_content = ""
        self.__render()

    def get_rendered_config(self):
        return self.__syslog_ng_config_content

    def __render(self, re_create_config=None):

        version = self.__syslog_ng_config["version"]
        includes = self.__syslog_ng_config["includes"]
        global_options = self.__syslog_ng_config["global_options"]
        preamble = self.__syslog_ng_config["preamble"]
        templates = self.__syslog_ng_config["templates"]
        statement_groups = self.__syslog_ng_config["statement_groups"]
        logpath_groups = self.__syslog_ng_config["logpath_groups"]

        config = ""

        if version:
            config += render_version(version)
        if includes:
            config += render_includes(includes)
        if global_options:
            config += render_global_options(global_options)
        if preamble:
            config += preamble
        if templates:
            config += render_templates(templates)
        if statement_groups:
            config += render_statement_groups(statement_groups)
        if logpath_groups:
            config += render_logpath_groups(logpath_groups)

        if re_create_config:
            self.__syslog_ng_config_content = config
        else:
            self.__syslog_ng_config_content += config



#!/usr/bin/env python
#############################################################################
# Copyright (c) 2015-2018 Balabit
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import time


class LogMessage(object):
    def __init__(self):
        self.priority_value = "38"
        self.timestamp_value = time.time()
        self.bsd_timestamp_value = "Feb 11 21:27:22"
        self.iso_timestamp_value = "2019-02-11T21:27:22+01:00"
        self.hostname_value = "testhost"
        self.program_value = "testprogram"
        self.pid_value = "9999"
        self.message_value = "test message"

    def priority(self, pri):
        self.priority_value = pri
        return self

    def remove_priority(self):
        self.priority_value = ""
        return self

    def timestamp(self, timestamp):
        self.timestamp_value = timestamp
        return self

    def hostname(self, hostname):
        self.hostname_value = hostname
        return self

    def program(self, program):
        self.program_value = program
        return self

    def pid(self, pid):
        self.pid_value = pid
        return self

    def message(self, message):
        self.message_value = message
        return self
